      program   driver
c     ------------------------------------------------------------------
c
c     driver program for
c     gmsatl: A GRASP fortran subroutine for Weighted MAX-SAT
c     large memeory version
c     Version : 2.10
c
c     Authors:  Mauricio G.C. Resende   (mgcr@research.att.com      )
c               Leonidas S.V. Pitsoulis (leonidas@deming.ise.ufl.edu)
c               Panos M. Pardalos       (pardalos@ufl.edu)
c
c     This code applies a GRASP to find high-valued truth assignments
c     in a weighted MAXSAT problem.  The GRASP is described in
c     M.G.C. Resende, L.S. Pitsoulis, and P.M. Pardalos, "Approximate 
c     solution of weighted MAX-SAT problems using GRASP," in DIMACS
c     Series in Discrete Mathematics and Theoretical Computer Science, 
c     vol. 35, pp. 393-405, 1997.
c
c     This driver program calls the following:
c
c           readp  : to input the MAX-SAT instance
c           errmsg : if an error condition is found
c           iniprt : to print out parameters and problem dimensions
c           gmsatl : to run GRASP (large memory version)
c           outsol : to print out the results
c        
c     The GRASP subroutine gmsatl can be called alone, as long as
c     proper parameters and arrays are passed to it.
c
c     The GRASP subroutine gmsatl takes as input 5 parameters:
c
c           iout   : output device number (standard output = 6)
c           niter  : maximum number of GRASP iterations (niter>=1)
c           seed   : seed for random number generator (1<=seed<=2^31-1)
c           prttyp : output control in gmsatl (0=silent;1=only 
c                    iterations with improvements;2=all iterations)
c           look4  : stop GRASP if assignment with weight at least as
c                    large as look4 is found
c
c     Array dimensions defined in calling program are passed to gmsatl:
c
c           nmax   - maximum number of variables
c           mmax   - maximum number of clauses
c           lmax   - maximum number of literals
c
c     The MAX-SAT instance is passed to gmsatl as 
c
c        3 integer scalars:
c
c           n      - number of variables (<=nmax)
c
c           m      - number of clauses (<=nmax)
c
c           numlit - number of literals (<=lmax)
c
c        and 3 integer arrays:
c
c           headc  - headc(i) is the position in array lit where the 
c                    first literal of clause i is stored; headc(m+1)
c                    points to first position in array lit after last 
c                    literal of clause m (dimension: m+1)
c
c           lit    - array of variable of literal (<0=negated;
c                    >0=unnegated) (dimension: numlit)
c
c           w      - array element w(i) is weight of clause i 
c                    (dimension: mmax)
c
c     The solution is passed back from gmsatl in
c
c        3 integer scalars:
c
c           bestv  - weight of best assignment
c           bests  - seed at start of iteration that produced best 
c                    assignment
c           iter   - number of GRASP iterations carried out
c
c        and 1 integer array:
c
c           xopt   - best truth assignment (xopt(i)=0 => false; 
c                    =1 => true)
c
c     Several auxiliary integer arrays are needed by gmsatl:
c
c           clause - dimension: m
c           gainx  - dimension: n
c           headm  - dimension: n
c           headp  - dimension: n
c           minus  - dimension: n
c           minus0 - dimension: n
c           nextm  - dimension: numlit
c           nextp  - dimension: numlit
c           plus   - dimension: n
c           plus0  - dimension: n
c           rcl    - dimension: n*2
c           satcl  - dimension: m
c           x      - dimension: n
c           xptr   - dimension: n+1
c           xx     - dimension: n*n
c
c     ------------------------------------------------------------------
      implicit none

c     ------------------------------------------------------------------
c
c     Variable definitions: integer scalars
c
c        lmax           - maximum number of literals
c        mmax           - maximum number of clauses
c        mp1max         - mmax + 1
c        nmax           - maximum number of variables
c        np1max         - nmax + 1
c        nt2max         - nmax * 2
c        ntnmax         - nmax * nmax
c
c     ------------------------------------------------------------------
      integer  lmax,   mmax,   nmax
      parameter ( nmax=100,mmax=900,lmax=10000 )

      integer  mp1max, np1max, nt2max, ntnmax
      parameter ( mp1max=mmax+1, np1max=nmax+1, nt2max=2*nmax,
     +            ntnmax=nmax*nmax)

      integer  maxelt
      parameter ( maxelt=50)
c     ------------------------------------------------------------------
c     I/O parameters
c
c        in             - input device number
c        iout           - output device number
c
c     ------------------------------------------------------------------
      integer  in,     iout
      parameter (in=5,iout=6)

c     ------------------------------------------------------------------      
c
c     Variable definitions: integer scalars
c
c        bestv          - value of best truth assignment
c        bests          - seed at start of iteration of best solution
c        errcnd         - error condition
c        iter           - iteration counter
c        look4          - stop GRASP if assignment with cost look4 is 
c                         found
c        m              - number of clauses
c        n              - number of variables
c        niter          - maxmimum number of GRASP iterations
c        numlit         - number of literals
c        prttyp         - printer option (0=silent, 1=improvements only,
c                                         2=all iterations)
c        seed           - random number generator seed
c
c     ------------------------------------------------------------------
      integer  bestv,  bests,  errcnd, iter,   look4,  m,      n,
     +         niter,  numlit, prttyp, seed

c     ------------------------------------------------------------------
c
c     Variable definitions: integer arrays
c
c        clause(mlit)   - clause which literal belongs to
c        gainx(nmax)    - gain achieved by flipping assignment
c        headc(mp1max)  - pointer to first literal in clause
c        headm(nmax)    - pointer to first negated literal of variable
c        headp(nmax)    - pointer to first unnegated literal of variable
c        lit(mlit)      - variable of literal (>0=unnegated, <0=negated)
c        minus(nmax)    - clause in which variable occurs as a negated
c                         literal of variable (updated)
c        minus0(nmax)   - clause in which variable occurs as a negated
c                         literal of variable (original)
c        nextm(mlit)    - pointer to next minus clause
c        nextp(mlit)    - pointer to next plus clause
c        plus(nmax)     - clause in which variable occurs as a unnegated
c                         literal of variable (updated)
c        plus0(nmax)    - clause in which variable occurs as a unnegated
c                         literal of variable (original)
c        rcl(nt2max)    - elements of RCL
c        satcl(mmax)    - indicator of satisfiable clause
c        w(mmax)        - clause weight
c        x(nmax)        - assignment
c        xopt(nmax)     - best assignment
c        xptr(np1max)   - pointer to first element of variable in 
c                         relationship list
c        xx(ntnmax)     - relationship list
c
c     ------------------------------------------------------------------
      integer  clause(lmax),   gainx(nmax),    headc(mp1max), 
     +         headm(nmax),    headp(nmax),    lit(lmax),
     +         minus(nmax),    minus0(nmax),   nextm(lmax),
     +         nextp(lmax),    plus(nmax),     plus0(nmax),
     +         rcl(nt2max),    satcl(mmax),    w(mmax),
     +         x(nmax),        xopt(nmax),     xptr(np1max),
     +         xx(ntnmax)

      integer  seede, elite(maxelt*(nmax+1)), symdif(nmax), msymdf,
     +         mxdpth, ypr(nmax), optxpr(nmax)
c     ------------------------------------------------------------------
c     Read problem data. 
c
c     Problem dimension (n,m) is read and numlit is computed.  The 
c     clauses are placed in (headc,lit,w).  If lmax, mmax, or nmax, are
c     not sufficient, an error condition (errcnd > 0) is returned. An
c     error condition errcnd=0 indicates successful execution.
c
c     ------------------------------------------------------------------
      read (5,*) seed,seede
      read (5,*) look4
      call readp( errcnd, headc,  in,     lit,    lmax,   m,      mmax,
     +            n,      nmax,   numlit, w )

c     ------------------------------------------------------------------
c     If error occurred in readp, halt execution after error message.
c     ------------------------------------------------------------------
      if ( errcnd .gt. 0) then
	 call errmsg( errcnd, iout )
         stop
      endif

c     ------------------------------------------------------------------
c
c     Set GRASP and program parameters.
c
c        look4          - stop GRASP if assignment with cost look4 is 
c                         found
c
c        niter          - maxmimum number of GRASP iterations
c
c        prttyp         - printer option (0=silent, 1=improvements only,
c                                         2=all iterations)
c
c        seed           - initial random number generator seed
c
c     ------------------------------------------------------------------
c     look4=2147483647
c     look4=800000
      niter=200000
      prttyp=1
c     seed=1778090629
c     seede=270001
      msymdf = 20
      mxdpth = 100

c     ------------------------------------------------------------------
c     Print problem and pre-run statistics.
c     ------------------------------------------------------------------
      call iniprt( iout,   look4,  m,      n,      niter,  numlit, 
     +             prttyp, seed )

c     ------------------------------------------------------------------
c     Call subroutine gmsatl to find approximate solution to weighted
c     MAXSAT problem.
c     ------------------------------------------------------------------
      call gmsatl( bestv,  bests,  clause, gainx,  headc,  headm,  
     +             headp,  iout,   iter,   lit,    look4,  lmax,
     +             m,      minus,  minus0, mmax,   n,      nextm,
     +             nextp,  niter,  nmax,   numlit, plus,   plus0,  
     +             prttyp, rcl,    satcl,  seed,   w,      x,      
     +             xopt,   xptr,   xx,     maxelt, elite,  seede,
     +             symdif, msymdf, mxdpth, ypr, optxpr ) 

c     ------------------------------------------------------------------
c     Print out best solution found or error message.
c     ------------------------------------------------------------------
      call errmsg( errcnd, iout )
      if ( errcnd .eq. 0) then
	      call outsol( bestv,  bests,  iter,   n,      xopt)
      endif

c     ------------------------------------------------------------------
      stop
      end








      subroutine readp(errcnd, headc,  in,     lit,    lmax,   m,
     +                 mmax,   n,      nmax,   numlit, w)
c     -----------------------------------------------------------------
c
c     readp: Reads MAXSAT instance and loads into (m,n,lit,headc,w).
c
c     -----------------------------------------------------------------
      implicit none

c     ------------------------------------------------------------------
c
c     Integer input parameters:
c
c        in             - input device number
c        lmax           - maximum number of literals
c        m              - number of clauses
c        mmax           - maximum number of clauses
c        n              - number of variables
c        nmax           - maximum number of variables
c
c     ------------------------------------------------------------------
      integer  in,     lmax,   m,      mmax,   n,      nmax

c     ------------------------------------------------------------------
c
c     Integer output parameters:
c
c        errcnd         - error condition
c        numlit         - number of literals
c
c     ------------------------------------------------------------------
      integer  errcnd, numlit

c     ------------------------------------------------------------------
c
c     Integer local parameters:
c
c        i              - do loop
c        j              - do loop
c        litcl          - number of literals in a clause
c
c     ------------------------------------------------------------------
      integer  i,      j,      litcl

c     ------------------------------------------------------------------
c
c     Integer output arrays:
c
c        headc(mmax)    - pointer to first literal in clause
c        lit(mlit)      - variable of literal (>0=unnegated, <0=negated)
c        w(mmax)        - clause weight
c
c     ------------------------------------------------------------------
      integer headc(mmax),    lit(lmax),      w(mmax)

c     ------------------------------------------------------------------
c     Read problem dimensions: (n,m)
c     ------------------------------------------------------------------
      read (in,*) n,m

c     ------------------------------------------------------------------
c     Test if enough storage has been allocated.
c     ------------------------------------------------------------------
      if (n.gt.nmax) then
           errcnd=1
           return
      endif
      if (m.gt.mmax) then
           errcnd=2
           return
      endif

c     ------------------------------------------------------------------
c     Read clauses, one at a time.  Count number of litreals (numlit).
c     Load problem into arrays (headc, lit, w).
c     ------------------------------------------------------------------
      numlit=0
      do 10 i = 1,m
           headc(i)=numlit+1
           read (in,*) litcl, w(i), (lit(numlit+j),j=1,litcl)
           numlit=numlit+litcl
c          -------------------------------------------------------------
c          Test if enough storage has been allocated.
c          -------------------------------------------------------------
           if (numlit.gt.lmax) then
                errcnd=3
                return
           endif
10    continue
      headc(m+1)=numlit+1
c     ------------------------------------------------------------------------
      return
      end







      subroutine errmsg( errcnd, iout )
c     ------------------------------------------------------------------
c
c     errmsg: Prints error message
c
c     ------------------------------------------------------------------
      implicit none

c     ------------------------------------------------------------------
c
c     Integer input parameters:
c
c     ------------------------------------------------------------------
c
c         errcnd         - error condition
c         iout           - output device number
c
c     ------------------------------------------------------------------
      integer errcnd,         iout

c     ------------------------------------------------------------------
      if (errcnd .eq. 0) then
           write(iout,10)
10         format(/' Execution terminated with no error. ')
           return
      endif

      if (errcnd .eq. 1) then
           write(iout,20)
20         format(/' Execution aborted.  Error in n (< 1 or > nmax) ')
           return
      endif

      if (errcnd .eq. 2) then
           write(iout,30)
30         format(/' Execution aborted.  Error in m (< 1 or > mmax) ')
           return
      endif

      if (errcnd .eq. 3) then
           write(iout,40)
40         format(/' Execution aborted.  '
     +            ' Error in numlit (< 1 or > lmax) ')
           return
      endif

      if (errcnd .eq. 4) then
           write(iout,50)
50         format(/' Execution aborted. '
     +            ' Error in w(*) (< 0 or > 2147483647) ')
           return
      endif

      if (errcnd .eq. 5) then
           write(iout,60)
60         format(/' Execution aborted. '
     +            ' Error in look4 (< 1 or > 2147483647) ')
           return
      endif

      if (errcnd .eq. 6) then
           write(iout,70)
70         format(/' Execution aborted. '
     +            ' Error in niter (< 1 or > 2147483647) ')
           return
      endif

      if (errcnd .eq. 7) then
           write(iout,80)
80         format(/' Execution aborted. '
     +            ' Error in prttyp (not equal to 0, 1, or 2) ')
           return
      endif

      if (errcnd .eq. 8) then
           write(iout,90)
90         format(/' Execution aborted. '
     +            ' Error in iseed (< 1 or > 2147483647) ')
           return
      endif

      if (errcnd .eq. 9) then
           write(iout,100)
100        format(/' Execution aborted. '
     +            ' Error in array headc (headc(i) > lmax)')
           return
      endif

      if (errcnd .eq. 10) then
           write(iout,110)
110        format(/' Execution aborted. '
     +            ' Error in array headc (headc(i) = headc(i+1))')
           return
      endif

      if (errcnd .eq. 11) then
           write(iout,120)
120        format(/' Execution aborted. '
     +            ' Error in array headc (headc(i) > headc(i+1))')
           return
      endif

      if (errcnd .eq. 12) then
           write(iout,130)
130        format(/' Execution aborted. '
     +            ' Error in array lit (< -n or > n)')
           return
      endif

c     ------------------------------------------------------------------
      return
      end








      subroutine iniprt( iout,   look4,  m,      n,      niter,  numlit,
     +                   prttyp, seed )
c     ------------------------------------------------------------------
c
c     iniprt: Writes summary of input data.
c
c     ------------------------------------------------------------------
      implicit none

c     ------------------------------------------------------------------
c
c     Integer input parameters:
c
c        iout           - output device number
c        look4          - stop GRASP if assignment with cost look4 is 
c                         found
c        m              - number of clauses
c        n              - number of variables
c        niter          - maxmimum number of GRASP iterations
c        numlit         - number of literals
c        prttyp         - printer option (0=silent, 1=improvements only,
c                                         2=all iterations)
c        seed           - random number generator seed
c
c     ------------------------------------------------------------------
      integer  iout,   look4,  m,      n,      niter,  numlit, prttyp,
     +         seed

c     ------------------------------------------------------------------
      write(iout,10)
10    format ('        GRASP/MAX-SAT: ')
      write(iout,20)
20    format ('                       ')
      write(iout,30) n
30    format (' number of variables : ',i10)
      write(iout,40) m
40    format ('   number of clauses : ',i10)
      write(iout,50) numlit
50    format ('  number of literals : ',i10)
      write(iout,60) niter
60    format ('    iteration limit  : ',i10)
      write(iout,70) seed
70    format ('        initial seed : ',i10)
      write(iout,80) prttyp
80    format ('        print option : ',i10)
      write(iout,90) look4
90    format ('            look for : ',i10)
      write(iout,91) niter
91    format ('            niter    : ',i10)
c     ------------------------------------------------------------------
      return
      end








      subroutine outsol( bestv,  bests,  iter,   n,      xopt )
c     --------------------------------------------------------------
c
c     outsol: output best solution found
c
c     --------------------------------------------------------------
      implicit none
      integer  bestv,  bests,  i,      iter,   n
      integer  xopt(n)
c     --------------------------------------------------------------
      print *,'                        '
      print *,'         GRASP/output : '
      print *,'                        '
      print *,' number of iterations : ',iter
      print *,' seed  best iteration : ',bests
      print *,'        best solution : ',bestv
      print *,'                        '

      write(6,10) (xopt(i),i=1,n)
10    format(25h      truth assignment : ,22i2)
      print *,'              '
c     --------------------------------------------------------------
      return
      end
