      subroutine gmsatl( bestv,  bests,  clause, gainx,  headc,  headm,  
     +                   headp,  iout,   iter,   lit,    look4,  lmax,
     +                   m,      minus,  minus0, mmax,   n,      nextm,
     +                   nextp,  niter,  nmax,   numlit, plus,   plus0,
     +                   prttyp, rcl,    satcl,  seed,   w,      x,
     +                   xopt,   xptr,   xx,     maxelt, elite,  seede,
     +                   symdif, msymdf, mxdpth, ypr, optxpr) 
c     ------------------------------------------------------------------
c
c     gmsatl: A GRASP fortran subroutine for Weighted MAX-SAT
c     large memory version
c     Version : 2.10
c
c     Authors:  Mauricio G.C. Resende   (mgcr@research.att.com      )
c               Leonidas S.V. Pitsoulis (leonidas@deming.ise.ufl.edu)
c               Panos M. Pardalos       (pardalos@ufl.edu)
c
c     This code applies a GRASP to find high-valued truth assignments
c     in a weighted MAXSAT problem.  The GRASP is described in
c     M.G.C. Resende, L.S. Pitsoulis, and P.M. Pardalos, "Approximate 
c     solution of weighted MAX-SAT problems using GRASP," in DIMACS
c     Series in Discrete Mathematics and Theoretical Computer Science, 
c     vol. 35, pp. 393-405, 1997.
c
c     Subroutine gmsatl requires proper parameters and arrays are 
c     passed to it from the calling program.
c
c     The subroutine takes as input 5 parameters:
c
c           iout   : output device number (standard output = 6)
c           niter  : maximum number of GRASP iterations (niter>=1)
c           seed   : seed for random number generator (1<=seed<=2^31-1)
c           prttyp : output control in gmsatl (0=silent;1=only 
c                    iterations with improvements;2=all iterations)
c           look4  : stop GRASP if assignment with weight at least as
c                    large as look4 is found
c
c     The array dimensions defined in calling program are passed 
c     to gmsatl:
c
c           nmax   - maximum number of variables
c           mmax   - maximum number of clauses
c           lmax   - maximum number of literals
c
c     The MAX-SAT instance is passed to subroutine gmsatl as:
c
c        3 integer scalars:
c
c           n      - number of variables (<=nmax)
c
c           m      - number of clauses (<=nmax)
c
c           numlit - number of literals (<=lmax)
c
c        and 3 integer arrays:
c
c           headc  - headc(i) is the position in array lit where the 
c                    first literal of clause i is stored; headc(m+1)
c                    points to first position in array lit after last 
c                    literal of clause m (dimension: m+1)
c
c           lit    - array of variable of literal (<0=negated;
c                    >0=unnegated) (dimension: numlit)
c
c           w      - array element w(i) is weight of clause i 
c                    (dimension: mmax)
c
c     The solution is passed back from gmsatl to the calling program in:
c
c        3 integer scalars:
c
c           bestv  - weight of best assignment
c           bests  - seed at start of iteration that produced best 
c                    assignment
c           iter   - number of GRASP iterations carried out
c
c        and 1 integer array:
c
c           xopt   - best truth assignment (xopt(i)=0 => false; 
c                    =1 => true)
c
c     Several auxiliary integer arrays are needed by subroutine gmsatl:
c
c           clause - dimension: m
c           gainx  - dimension: n
c           headm  - dimension: n
c           headp  - dimension: n
c           minus  - dimension: n
c           minus0 - dimension: n
c           nextm  - dimension: numlit
c           nextp  - dimension: numlit
c           plus   - dimension: n
c           plus0  - dimension: n
c           rcl    - dimension: n*2
c           satcl  - dimension: m
c           x      - dimension: n
c           xptr   - dimension: n+1
c           xx     - dimension: n*n
c
c     ------------------------------------------------------------------
      implicit none

c     ------------------------------------------------------------------
c
c     Integer input parameters:
c
c        iout           - output device number
c        look4          - stop GRASP if assignment with cost look4 is 
c                         found
c        m              - number of clauses
c        lmax           - maximum number of literals
c        mmax           - maximum number of clauses
c        n              - number of variables
c        niter          - maxmimum number of GRASP iterations
c        nmax           - maximum number of variables
c        numlit         - number of literals
c        prttyp         - printer option (0=silent, 1=improvements only,
c                                         2=all iterations)
c        seed           - random number generator seed
c
c     ------------------------------------------------------------------
      integer iout,   lmax,   look4,  m,      mmax,   n,      niter,
     +        nmax,   numlit, prttyp, seed
     
c     ------------------------------------------------------------------
c
c     Integer output parameters:
c
c        bestv          - value of best truth assignment
c        bests          - seed at start of iteration of best solution
c        errcnd         - error condition
c        iter           - iteration counter
c
c     ------------------------------------------------------------------
      integer  bestv,  bests,  errcnd, iter

c     ------------------------------------------------------------------
c
c     Integer local parameters:
c
c        cobjv          - value of truth assignment in construction 
c                         phase
c        i              - do loop counter
c        mp1            - m + 1
c        np1            - n + 1
c        nrcl           - size of RCL
c        nt2            - n * 2
c        ntn            - n * n
c        pkey           - indicates whether improvement was found
c        seed0          - seed at start of GRASP iteration
c        sumw           - sum of clause weights
c        wsat           - weight of truth assignment
c
c     ------------------------------------------------------------------
      integer  cobjv,  i,      mp1,    np1,    nrcl,   nt2,    ntn,
     +         pkey,   seed0,  sumw,   wsat

c     ------------------------------------------------------------------
c
c     Real local parameters:
c
c        alpha          - RCL parameter
c
c     ------------------------------------------------------------------
      real     alpha

c     ------------------------------------------------------------------
c
c     Integer input arrays:
c
c        headc(m+1)     - pointer to first literal in clause
c        lit(lmax)      - variable of literal (>0=unnegated, <0=negated)
c        w(m)           - clause weight
c
c     ------------------------------------------------------------------
      integer headc(m+1),     lit(lmax),      w(m)

c     ------------------------------------------------------------------
c
c     Integer output array:
c
c        xopt(n)        - best assignment
c
c     ------------------------------------------------------------------
      integer xopt(n)

c     ------------------------------------------------------------------
c
c     Integer local arrays:
c
c        clause(lmax)   - clause which literal belongs to
c        gainx(n)       - gain achieved by flipping assignment
c        headm(n)       - pointer to first negated literal of variable
c        headp(n)       - pointer to first unnegated literal of variable
c        minus(n)       - clause in which variable occurs as a negated
c                         literal of variable (updated)
c        minus0(n)      - clause in which variable occurs as a negated
c                         literal of variable (original)
c        nextm(lmax)    - pointer to next minus clause
c        nextp(lmax)    - pointer to next plus clause
c        plus(n)        - clause in which variable occurs as a unnegated
c                         literal of variable (updated)
c        plus0(n)       - clause in which variable occurs as a unnegated
c                         literal of variable (original)
c        rcl(nt2max)    - elements of RCL
c        satcl(m)       - indicator of satisfiable clause
c        x(n)           - assignment
c        xptr(n+1)      - pointer to first element of clause in 
c                         relationship list
c        xx(ntnmax)     - relationship list
c
c     ------------------------------------------------------------------
      integer clause(lmax),   gainx(n),       headm(n),
     +        headp(n),       minus(n),       minus0(n),
     +        nextm(lmax),    nextp(lmax),    plus(n),
     +        plus0(n),       rcl(2*n),       satcl(m),
     +        x(n),           xptr(n+1),      xx(n*n)

c     ------------------------------------------------------------------
c     elite variables and arrays
c     ------------------------------------------------------------------
      integer maxelt, elite(maxelt*(n+1)), nelite, seede, symdif(n), 
     r        msymdf, mxdpth, ypr(n), bstelt, wstelt, optxpr(n),bestpr,
     +        lssol

      real cput(2),btime,ttime,etime
c     ------------------------------------------------------------------
c     Check if memory dimension is sufficient, if input graph
c     is consistent, and if algorithm parameters are allowable.
c     Return control to calling program if error is found.
c     ------------------------------------------------------------------
      errcnd=0
      call chkinp( errcnd, headc,  lit,    lmax,   look4,  m,     niter,
     +             mmax,   n,      nmax,   numlit, prttyp, seed,   w )
 
      if (errcnd.gt.0) return
	
c     ------------------------------------------------------------------
c     Define array dimensions mp1, np1, ntn, and nt2.
c     ------------------------------------------------------------------
      mp1=m+1
      np1=n+1
      ntn=n*n
      nt2=n*2

c     ------------------------------------------------------------------
c     Create linked list data structures.
c     ------------------------------------------------------------------
      call mkdsl( clause, headc,  headm,  headp,  lit,    m,      
     +            minus0, mp1,    n,      nextm,  nextp,  np1,    ntn,
     +            numlit, plus0,  sumw,   w,      xptr,   xx )

c     ------------------------------------------------------------------
c     Initialize value of objective function that satisfies search.
c     ------------------------------------------------------------------
c     look4=min0(look4,sumw)

c     ------------------------------------------------------------------
c     Initialize value of best solution found.
c     ------------------------------------------------------------------
      bestv=0

c     ------------------------------------------------------------------
c     Initialize elite set.
c     ------------------------------------------------------------------
      nelite=0
      bstelt=0
      wstelt=2147483647
      ttime=etime(cput)
      btime=cput(1)

c     ------------------------------------------------------------------
c     GRASP ITERATIONS....
c     ------------------------------------------------------------------
      do 40 iter = 1,niter
c          -------------------------------------------------------------
c          Remember seed at start of GRASP iteration.
c          -------------------------------------------------------------
           seed0=seed
c          -------------------------------------------------------------
c          Computer RCL parameter alpha for this iteration.
c          -------------------------------------------------------------
           call randp(seed)
           alpha= float(seed)/2147483647
c          -------------------------------------------------------------
c          Create initial RCL (restricted candidate list).
c          -------------------------------------------------------------
           call mkrcl( alpha,  minus0, n,      nrcl,   nt2,    plus0,
     +                 rcl )

c          -------------------------------------------------------------
c          Copy original data structure into (minus, plus)
c          -------------------------------------------------------------
           call copyi4( n,      minus0, minus )
           call copyi4( n,      plus0,  plus )

c          -------------------------------------------------------------
c          GRASP construction phase.
c          -------------------------------------------------------------
           call build( alpha,  clause, headc,  headm,  headp,  lit,
     +                 m,      minus,  mp1,    n,      nextm,  nextp,
     +                 nrcl,   nt2,    numlit, plus,   rcl,    satcl,
     +                 seed,   w,      wsat,   x )

c          -------------------------------------------------------------
c          Save GRASP construction phase solution.
c          -------------------------------------------------------------
           cobjv=wsat

c          -------------------------------------------------------------
c          GRASP local search phase.
c          -------------------------------------------------------------
           call locall( clause, gainx,  headc,  headm,  headp,  lit,
     +                  m,      mp1,    n,      nextm,  nextp,  np1,
     +                  ntn,    numlit, w,      wsat,   x,      xptr,
     +                  xx )
           lssol=wsat

c          -------------------------------------------------------------
c          Path-relinking phase
c          -------------------------------------------------------------
           call pr( clause, gainx,  headc,  headm,  headp,  lit,
     +              m,      mp1,    n,      nextm,  nextp,  np1,
     +              ntn,    numlit, w,      wsat,   x,      xptr,
     +              xx, 
     +              nelite, elite, maxelt, seede, symdif, msymdf,
     +              mxdpth, ypr, bstelt, wstelt, optxpr, bestpr)

c          -------------------------------------------------------------
c          Save solution if best so far.
c          -------------------------------------------------------------
           pkey=0
           if (bestpr.gt.bestv) then
                call savsol( bestv,  bests,  n,      seed0,  bestpr,  
     +                       optxpr, xopt )
                if (prttyp .ne. 0) then
		     ttime=etime(cput)
		     ttime=cput(1)-btime
                     pkey=1
                     if (iter .eq. 1) then
c                        write (iout,10)
c10                       format(27h      iter alpha  ph-1-soln
c    +                      40h  ph-2-soln  pr-soln best-soln tot_time)
                     endif
                     print *,"ITR:", iter,alpha,cobjv,lssol,bestpr,
     +                           bestv,ttime
c20                   format(i10,1h ,f5.2,1h ,i10,1h ,i10,1h ,i10,1h ,i10,
c    +                      f10.2,16h <<<<improvement)
                endif
c               --------------------------------------------------------
c               If satisfiable assignment is found, return.
c               --------------------------------------------------------
                if (bestv .ge. look4) goto 50
           endif

c          -------------------------------------------------------------
c          Print summary report.
c          -------------------------------------------------------------

          if (prttyp.eq.2 .and. pkey .eq. 0) then
                write(6,30) iter,alpha,cobjv,wsat,bestv
30              format(i10,1h ,f5.2,1h ,i10,1h ,i10,1h ,i10)
           endif
40    continue
c     ------------------------------------------------------------------
c     Adjust iteration counter.
c     ------------------------------------------------------------------
      iter=iter-1

c     --------------------------------------------------------------
c     Transform truth assignment -1 to 0.
c     --------------------------------------------------------------
50    do 60 i=1,n
           if (xopt(i).eq.-1) xopt(i)=0
60    continue

c     ------------------------------------------------------------------
      return
      end








      subroutine chkinp( errcnd, headc,  lit,    lmax,   look4,  m,
     +                   niter,  mmax,   n,      nmax,   numlit, prttyp,
     +                   seed,   w )
c     ------------------------------------------------------------------
c
c     chkinp: Check input problem instance.
c
c     ------------------------------------------------------------------
      implicit none

c     ------------------------------------------------------------------
c
c     Integer input parameters:
c
c         look4          - look for planar subgraph with look4 arcs
c         lmax           - max number of literals
c         m              - number of arc
c         mmax           - max value of dimension m
c         niter          - maximum number of iterations
c         n              - number of nodes
c         nmax           - max value of dimension n
c         numlit         - number of literals
c         prttyp         - type of reporting (0=silent,1=improvement,
c                          2=all iterations)
c         seed           - seed for pseudo random number generator
c
c     ------------------------------------------------------------------

      integer bigint, i,      k,      lmax,   look4,  m,      mmax,
     +        n,      niter,  nmax,   numlit, prttyp, seed
 
      parameter (bigint=2147483647)

c     ------------------------------------------------------------------
c
c     Integer output parameter:
c
c         errcnd         - error condition
c
c     ------------------------------------------------------------------
      integer errcnd

c     ------------------------------------------------------------------
c
c     Integer input arrays:
c
c         headc(mmax)     - pointer to start of clause 
c         lit(lmax)       - literals in clause
c         w(mmax)         - weight of clause
c
c     -----------------------------------------------------------------
      integer headc(mmax),  lit(lmax),      w(mmax)

c     ------------------------------------------------------------------
c     Check n, m, and  numlit.
c     ------------------------------------------------------------------
      if (n.lt.1 .or. n.gt.nmax) then
           errcnd=1
           return
      endif

      if (m.lt.1 .or. m.gt.mmax) then
           errcnd=2
           return
      endif

      if (numlit.lt.1 .or. numlit.gt.lmax) then
           errcnd=3
           return
      endif

c     ------------------------------------------------------------------
c     Check GRASP parameters: look4, niter, prttyp, seed.
c     ------------------------------------------------------------------
      if (look4 .lt. 0 .or. look4 .gt. bigint) then
           errcnd=5
           return
      endif

      if (niter .lt. 1 .or. niter .gt. bigint) then
           errcnd=6
           return
      endif

      if (.not.(prttyp.eq.0 .or. prttyp.eq.1 .or. prttyp.eq.2)) then
           errcnd=7
           return
      endif

      if (seed .lt. 1 .or. seed .gt. bigint) then
           errcnd=8
           return
      endif

c     ------------------------------------------------------------------
c     Check input data: headc, lit, and w.
c     ------------------------------------------------------------------
      do 20 i=1,m
           if (w(i) .lt. 0 .or. w(i) .gt. bigint) then
                errcnd=4
                return
           endif
           if (headc(i) .gt. lmax) then
                errcnd=9
                return
           endif
           if (headc(i) .eq. headc(i+1)) then
                errcnd=10
                return
           endif
           if (headc(i) .gt. headc(i+1)) then
                errcnd=11
                return
           endif

           do 10 k=headc(i),headc(i+1)-1
                if (lit(k) .lt. -n .or. lit(k) .gt. n) then
                     errcnd=12
                     return
                endif
10         continue
20    continue
c     ------------------------------------------------------------------
      return
      end








      subroutine mkdsl( clause, headc,  headm,  headp,  lit,    
     +                  m,      minus0, mp1,    n,      nextm,
     +                  nextp,  np1,    ntn,    numlit, plus0,
     +                  sumw,   w,      xptr,   xx )
c     -----------------------------------------------------------------
c
c     mkdsl: Builds data structures.
c
c     -----------------------------------------------------------------
      implicit none

c     ------------------------------------------------------------------
c
c     Integer input parameters:
c
c        m              - number of clauses
c        mp1            - m + 1
c        n              - number of variables
c        np1            - n + 1
c        ntn            - n * n
c        numlit         - number of literals
c
c     ------------------------------------------------------------------
      integer m,      mp1,    n,      np1,    ntn,    numlit

c     ------------------------------------------------------------------
c
c     Integer output parameters:
c
c        sumw           - sum of clause weights
c
c     ------------------------------------------------------------------
      integer sumw

c     ------------------------------------------------------------------
c
c     Integer local parameters:
c
c        i              - do loop parameter
c        index          - index of array
c        j              - do loop parameter
c        k              - do loop parameter
c        l              - literal counter
c        neg            - indicator
c        ptr            - pointer
c
c     ------------------------------------------------------------------
      integer i,      index,  j,      k,      l,      neg,    ptr

c     ------------------------------------------------------------------
c
c     Integer input arrays:
c
c        headc(mp1 )    - pointer to first literal in clause
c        lit(numlit)    - variable of literal (>0=unnegated, <0=negated)
c        w(m)           - clause weight
c
c     ------------------------------------------------------------------
      integer headc(mp1),     lit(numlit),    w(m)

c     ------------------------------------------------------------------
c
c     Integer output arrays:
c
c        clause(numlit) - clause which literal belongs to
c        headm(n)       - pointer to first negated literal of variable
c        headp(n)       - pointer to first unnegated literal of variable
c        minus0(n)      - clause in which variable occurs as a negated
c                         literal of variable (original)
c        nextm(numlit)  - pointer to next minus clause
c        nextp(numlit)  - pointer to next plus clause
c        plus0(n)       - clause in which variable occurs as a unnegated
c                         literal of variable (original)
c        xptr(np1)      - pointer to first element of clause in 
c                         relationship list
c        xx(ntn)        - relationship list
c
c     ------------------------------------------------------------------
      integer clause(numlit), headm(n),       headp(n),
     +        minus0(n),      nextm(numlit),  nextp(numlit),
     +        plus0(n),       xptr(np1),      xx(ntn)


c     ------------------------------------------------------------------
c     Set up clause array.
c     ------------------------------------------------------------------
      do 20 i=1,m
           do 10 j=headc(i),headc(i+1)
               clause(j)=i
10         continue
20    continue

c     ------------------------------------------------------------------
c     set up column linked list data structure
c     ------------------------------------------------------------------
      do 30 i=1,n
           headp(i)=0
           headm(i)=0
30    continue

      do 40 i=1,m
           plus0(i)=0
           minus0(i)=0
40    continue

      do 60 i=1,m
           do 50 j=headc(i),headc(i+1)-1
                index=iabs(lit(j))
                if (lit(j).lt.0) then
                     minus0(index)=minus0(index)+w(i)
                     nextm(j)=headm(index)
                     headm(index)=j
                else
                     plus0(index)=plus0(index)+w(i)
                     nextp(j)=headp(index)
                     headp(index)=j
                endif
50         continue
60    continue

c     -----------------------------------------------------------------
c     Set up the relationship list xx.
c     -----------------------------------------------------------------
      l=0
      do 130 i=1, n
           l=l+1
           xx(l)=i
           xptr(i)=l
           neg=0
           ptr=headp(i)
70         if (ptr.eq.0) goto 110
80              do 100 k=headc(clause(ptr)), headc(clause(ptr)+1)-1
                     index=iabs(lit(k))
                     do 90 j=xptr(i), l
                          if (index.eq.xx(j)) goto 100
90                   continue
                     l=l+1
                     xx(l)=index
100             continue
                if (neg.eq.1) then
                     ptr=nextm(ptr)
                     goto 120
                endif
                ptr=nextp(ptr)
                goto 70
110         continue
            ptr=headm(i)
            neg=1
120         if (ptr.eq.0) goto 130
            goto 80
130   continue
      xptr(n+1)=l+1

c     -----------------------------------------------------------------
c     Compute sum of weights.
c     -----------------------------------------------------------------
      sumw=0
      do 140 i=1,m
           sumw=sumw+w(i)
140   continue

c     -----------------------------------------------------------------
      return
      end








      subroutine randp(ix)
c     ------------------------------------------------------------------
c     randp: Portable pseudo-random number generator.
c            Reference: L. Schrage, "A More Portable Fortran
c            Random Number Generator", ACM Transactions on
c            Mathematical Software, June, 1979.
c     ------------------------------------------------------------------
      implicit none

      integer*4 a,p,ix,b15,b16,xhi,xalo,leftlo,fhi,k
      data a/16807/,b15/32768/,b16/65536/,p/2147483647/

c     ------------------------------------------------------------------
      xhi=ix/b16
      xalo=(ix-xhi*b16)*a
      leftlo=xalo/b16
      fhi=xhi*a+leftlo
      k=fhi/b15
      ix=(((xalo-leftlo*b16)-p)+(fhi-k*b15)*b16)+k
      if (ix.lt.0) ix=ix+p
c     ------------------------------------------------------------------
      return
      end








      subroutine mkrcl( alpha,  minus0, n,      nrcl,   nt2,    plus0,
     +                  rcl )
c     ------------------------------------------------------------------
c
c     mkrcl:  Builds restricted candidate list (RCL).
c
c     ------------------------------------------------------------------
      implicit none

c     ------------------------------------------------------------------
c
c     Integer input parameters:
c
c        n              - number of variables
c        nt2            - n * 2
c
c     ------------------------------------------------------------------
      integer  n,      nt2

c     ------------------------------------------------------------------
c
c     Integer output parameter:
c
c        nrcl           - number of elements in RCL
c
c     ------------------------------------------------------------------
      integer  nrcl

c     ------------------------------------------------------------------
c
c     Integer local parameters:
c
c        cutoff         - cut-off limit to be in RCL
c        i              - do loop parameter
c        maxsat         - maximum weight of newly satisfiable clauses
c        minsat         - minimum weight of newly satisfiable clauses
c
c     ------------------------------------------------------------------
      integer  cutoff, i,      maxsat, minsat

c     ------------------------------------------------------------------
c
c     Real input parameter:
c
c        alpha          - RCL parameter
c
c     ------------------------------------------------------------------
      real     alpha

c     ------------------------------------------------------------------
c
c     Integer input arrays:
c
c        minus(n)       - clause in which variable occurs as a negated
c                         literal of variable (updated)
c        plus(n)        - clause in which variable occurs as a unnegated
c                         literal of variable (updated)
c
c     ------------------------------------------------------------------
      integer  plus0(n),       minus0(n)

c     ------------------------------------------------------------------
c
c     Integer output array:
c
c        rcl(nt2)       - clause in which variable occurs as a negated
c                         literal of variable (updated)
c
c     ------------------------------------------------------------------
      integer  rcl(nt2)

c     ------------------------------------------------------------------
c     Initialize min and max weights.
c     ------------------------------------------------------------------
      minsat=2147483647
      maxsat=0
c     ------------------------------------------------------------------
c     Find max and min over all literals.
c     ------------------------------------------------------------------
      do 10 i=1,n
	   if (plus0(i).gt.maxsat) maxsat=plus0(i)
	   if (minus0(i).gt.maxsat) maxsat=minus0(i)
	   if (plus0(i).lt.minsat) minsat=plus0(i)
	   if (minus0(i).lt.minsat) minsat=minus0(i)
10    continue

c     ------------------------------------------------------------------
c     Compute cutoff for inclusion into RCL.
c     ------------------------------------------------------------------
      cutoff=minsat+ifix(alpha*(maxsat-minsat))

c     ------------------------------------------------------------------
c     Initialize size of RCL to empty.
c     ------------------------------------------------------------------
      nrcl=0

c     ------------------------------------------------------------------
c     Include into RCL all literal assignments that fall above cutoff.
c     ------------------------------------------------------------------
      do 20 i=1,n
         if (plus0(i).ge.cutoff) then
              nrcl=nrcl+1
              rcl(nrcl)=i
         endif
         if (minus0(i).ge.cutoff) then
              nrcl=nrcl+1
              rcl(nrcl)=-i
         endif
20    continue
c     ------------------------------------------------------------------
      return
      end







      subroutine build( alpha,  clause, headc,  headm,  headp,  lit,
     +                  m,      minus,  mp1,    n,      nextm,  nextp,
     +                  nrcl,   nt2,    numlit, plus,   rcl,    satcl,
     +                  seed,   w,      wsat,   x )
c     ------------------------------------------------------------------
c
c     build:  GRASP construction phase.  Builds a randomized greedy 
c             assignment.  
c
c             Assignment is returned in array x(*).
c
c     ------------------------------------------------------------------
      implicit none

c     ------------------------------------------------------------------
c
c     Integer input parameters:
c
c        m              - number of clauses
c        mp1            - m + 1
c        n              - number of variables
c        nt2            - n * 2
c        numlit         - number of literals
c        seed           - random number generator seed
c
c     ------------------------------------------------------------------
      integer  m,      mp1,    n     , nt2,    numlit, seed

c     ------------------------------------------------------------------
c
c     Integer output parameters:
c
c        wsat           - weight of assignment
c
c     ------------------------------------------------------------------
      integer wsat

c     ------------------------------------------------------------------
c
c     Integer local parameters:
c
c        index          - index of array
c        index2         - index of array
c        j              - do loop parameter
c        k              - do loop parameter
c        nrcl           - size of RCL
c        s              - random number between 1 and nrcl
c
c     ------------------------------------------------------------------
      integer  index,  index2, j,      k,      nrcl,   ptr,    s
c     ------------------------------------------------------------------
c
c     Real local parameters:
c
c        alpha          - RCL parameter
c
c     ------------------------------------------------------------------
      real    alpha

c     ------------------------------------------------------------------
c
c     Integer input arrays:
c
c        clause(numlit) - clause which literal belongs to
c        headc(mp1)     - pointer to first literal in clause
c        headm(n)       - pointer to first negated literal of variable
c        headp(n)       - pointer to first unnegated literal of variable
c        lit(numlit)    - variable of literal (>0=unnegated, <0=negated)
c        minus(n)       - clause in which variable occurs as a negated
c                         literal of variable (updated)
c        nextm(numlit)  - pointer to next minus clause
c        nextp(numlit)  - pointer to next plus clause
c        plus(n)        - clause in which variable occurs as a unnegated
c                         literal of variable (updated)
c        w(m)           - clause weight
c
c     ------------------------------------------------------------------
      integer clause(numlit), headc(mp1),     headm(m),
     +        headp(m),       lit(numlit),    minus(n),
     +        nextm(numlit),  nextp(numlit),  plus(n),
     +        w(m)

c     ------------------------------------------------------------------
c
c     Integer output array:
c
c        x(n)           - assignment
c
c     ------------------------------------------------------------------
      integer x(n)

c     ------------------------------------------------------------------
c
c     Integer local arrays:
c
c        rcl(nt2)       - elements of RCL
c        satcl(m)       - indicator of satisfiable clause
c
c     ------------------------------------------------------------------
      integer rcl(nt2),       satcl(m)


c     ------------------------------------------------------------------
c     Initialize satisfied clause indicator.
c     wsat.
c     ------------------------------------------------------------------
      do 10 j=1,m
         satcl(j)=0
10    continue
c     ------------------------------------------------------------------
c     Initialize solution vector x and satisfied clause weight counter 
c     wsat.
c     ------------------------------------------------------------------
      do 20 j=1,n
	   x(j)=0
20    continue
      wsat=0

c     ------------------------------------------------------------------
c     Main loop:  assignments 1,2,..,n are made.
c     ------------------------------------------------------------------
      do 50 j=1,n
c          -------------------------------------------------------------
c          If no candidate exists in RCL, construction phase is done.
c          -------------------------------------------------------------
	   if (nrcl .eq. 0) return

c          -------------------------------------------------------------
c          Select a candidate from RCL.
c          -------------------------------------------------------------
           call randp(seed)
           s=(seed/(2147483647/nrcl))+1
           index=iabs(rcl(s))
           
c          -------------------------------------------------------------
c          Scan literal & clause lists computing improvement in solution
c          with this assignment.
c          -------------------------------------------------------------
	   if (rcl(s) .gt. 0) then
                x(index)=1
                wsat=wsat+plus(index)
                ptr=headp(index)
           else 
                x(index)=-1
                wsat=wsat+minus(index)
                ptr=headm(index)
           endif

c          -------------------------------------------------------------
c          Check if done scanning literal.
c          -------------------------------------------------------------
30         if (ptr.ne.0) then
c               --------------------------------------------------------
c               Check if clause(ptr) is not yet satified.
c               --------------------------------------------------------
                if (satcl(clause(ptr)).ne.1) then
c                    ---------------------------------------------------
c                    Clause pointed to by ptr is not yet satisfied. 
c                    Scan clause and adjust potential weight for yet 
c                    unassigned literals in the clause.
c                    ---------------------------------------------------
                     do 40 k=headc(clause(ptr)),headc(clause(ptr)+1)-1
                        index2=iabs(lit(k))
                        if (x(index2).ne.0) goto 40
                        if (lit(k).gt.0) then
                           plus(index2)=plus(index2)-w(clause(ptr))
                        else
                           minus(index2)=minus(index2)-w(clause(ptr))
                        endif
40                   continue

c                    ---------------------------------------------------
c                    Mark clause(ptr) as satified.
c                    ---------------------------------------------------
                     satcl(clause(ptr))=1
                endif
c               --------------------------------------------------------
c               Move to next clause. 
c               --------------------------------------------------------
                if (rcl(s).gt.0) then
                   ptr=nextp(ptr)
                else
                   ptr=nextm(ptr)
                endif
		goto 30
           endif
c          -------------------------------------------------------------
c          Set literal index as unavailable for future assignment.
c          -------------------------------------------------------------
           plus(index)=0
           minus(index)=0

c          -------------------------------------------------------------
c          Update RCL taking into account new plus and minus values.
c          -------------------------------------------------------------
           call updrcl( alpha,  minus,  n,      nrcl,   nt2,    plus,
     +                  rcl,    x )
50    continue
c     ------------------------------------------------------------------
      return
      end








      subroutine updrcl( alpha,  minus,  n,      nrcl,   nt2,    plus,
     +                   rcl,    x )
c     ------------------------------------------------------------------
c
c     updrcl:  Builds updated restricted candidate list (RCL).
c
c     ------------------------------------------------------------------
      implicit none

c     ------------------------------------------------------------------
c
c     Integer input parameters:
c
c        n              - number of variables
c        nt2            - n * 2
c
c     ------------------------------------------------------------------
      integer  n,      nt2

c     ------------------------------------------------------------------
c
c     Real input parameter:
c
c        alpha          - RCL parameter
c
c     ------------------------------------------------------------------
      real     alpha

c     ------------------------------------------------------------------
c
c     Integer output parameter:
c
c        nrcl           - number of elements in RCL
c
c     ------------------------------------------------------------------
      integer  nrcl

c     ------------------------------------------------------------------
c
c     Integer local parameters:
c
c        cutoff         - cut-off limit to be in RCL
c        i              - do loop parameter
c        maxsat         - maximum weight of newly satisfiable clauses
c        minsat         - minimum weight of newly satisfiable clauses
c
c     ------------------------------------------------------------------
      integer  cutoff, i,      maxsat, minsat

c     ------------------------------------------------------------------
c
c     Integer input arrays:
c
c        minus(n)       - clause in which variable occurs as a negated
c                         literal of variable (updated)
c        plus(n)        - clause in which variable occurs as a unnegated
c                         literal of variable (updated)
c        x(n)           - assignment
c
c     ------------------------------------------------------------------
      integer  minus(n),       plus(n),        x(n)

c     ------------------------------------------------------------------
c
c     Integer output arrays:
c
c        rcl(nt2)       - clause in which variable occurs as a negated
c                         literal of variable (updated)
c
c     ------------------------------------------------------------------
      integer  rcl(nt2)

c     ------------------------------------------------------------------
c     Initialize min and max weights.
c     ------------------------------------------------------------------
      minsat=2147483647
      maxsat=0

c     ------------------------------------------------------------------
c     Find max and min over all yet unselected literals.
c     ------------------------------------------------------------------
      do 10 i=1,n
	   if (x(i).ne.0) goto 10
	   if (plus(i).gt.maxsat) maxsat=plus(i)
	   if (minus(i).gt.maxsat) maxsat=minus(i)
	   if (plus(i).lt.minsat) minsat=plus(i)
	   if (minus(i).lt.minsat) minsat=minus(i)
10    continue

c     ------------------------------------------------------------------
c     Initialize size of RCL to empty.
c     ------------------------------------------------------------------
      nrcl=0

c     ------------------------------------------------------------------
c     Compute cutoff for inclusion into RCL.
c     ------------------------------------------------------------------
      cutoff= minsat+ifix(alpha*(maxsat-minsat))

c     ------------------------------------------------------------------
c     Include into RCL all literal assignments that fall above cutoff.
c     ------------------------------------------------------------------
      do 20 i=1,n
	   if (x(i).ne.0) goto 20
	   if (plus(i).ge.cutoff) then
		nrcl=nrcl+1
		rcl(nrcl)=i
           endif
	   if (minus(i).ge.cutoff) then
		nrcl=nrcl+1
		rcl(nrcl)=-i
           endif
20    continue
c     ------------------------------------------------------------------
      return
      end








      subroutine locall( clause, gainx,  headc,  headm,  headp,  lit,
     +                   m,      mp1,    n,      nextm,  nextp,  np1,
     +                   ntn,    numlit, w,      wsat,   x,      xptr,
     +                   xx )
c     ------------------------------------------------------------------
c
c     locall: 1-exchange local optimization.   This search procedure
c             starts with the current constructed solution x and 
c             flips the assignment of the variable that makes the
c             most improvement in weight gain.  This is repeated until
c             no flip results in an improvement.
c
c     ------------------------------------------------------------------
      implicit none

c     ------------------------------------------------------------------
c
c     Integer input parameters:
c
c        m              - number of clauses
c        mp1            - m + 1
c        n              - number of variables
c        np1            - n + 1
c        ntn            - n * n
c        numlit         - number of literals
c
c     ------------------------------------------------------------------
      integer  m,      mp1,    n,      np1,    ntn,    numlit

c     ------------------------------------------------------------------
c
c     Integer input/output parameter:
c
c        wsat           - weight of assignment
c
c     ------------------------------------------------------------------
      integer  wsat

c     ------------------------------------------------------------------
c
c     Integer local parameters:
c
c        i              - literal index
c        ii             - do loop parameter
c        index          - literal index
c        k              - do loop parameter
c        l              - do loop parameter
c        lgain          - value of largest gain
c        li             - index of largest gain
c        p0             - xindex*y
c        p1             - -xindex*y
c        ptr             - pointer
c        ptrc            - clause(ptr)
c        scr0            - indicator
c        scr1            - indicator
c        smgain          - sum of gains
c        xindex          - x(index)
c        y               - indicator
c
c     ------------------------------------------------------------------
      integer  i,      ii,     index,  k,      l,      lgain,  li,
     +         p0,     p1,     ptr,    ptrc,   scr0,   scr1,    smgain,
     +         xindex, y

c     ------------------------------------------------------------------
c
c     Integer input arrays:
c
c        clause(mlit)   - clause which literal belongs to
c        headc(np1)     - pointer to first literal in clause
c        headm(nmax)    - pointer to first negated literal of variable
c        headp(nmax)    - pointer to first unnegated literal of variable
c        lit(mlit)      - variable of literal (>0=unnegated, <0=negated)
c        nextm(mlit)    - pointer to next minus clause
c        nextp(mlit)    - pointer to next plus clause
c        w(mmax)        - clause weight
c        xptr(mmax)     - pointer to first element of clause in 
c                         relationship list
c        xx(ntnmax)     - relationship list
c
c     ------------------------------------------------------------------
      integer  clause(numlit), headc(mp1),     headm(n),
     +         headp(n),       lit(numlit),    nextm(numlit),
     +         nextp(numlit),  w(m),           xptr(np1),
     +         xx(ntn)

c     ------------------------------------------------------------------
c
c     Integer input/output array:
c
c        x(nmax)        - assignment
c
c     ------------------------------------------------------------------
      integer  x(n)

c     ------------------------------------------------------------------
c
c     Integer local array:
c
c        gainx(nmax)    - gain achieved by flipping assignment
c
c     ------------------------------------------------------------------
      integer  gainx(n)

c     ------------------------------------------------------------------
      smgain=0
      lgain=0
      do 70 i=1,n
           gainx(i)=0
	   ptr=headp(i)
10         if(ptr.eq.0) goto 30
		scr0=0
		scr1=0
		ptrc=clause(ptr)
		do 20 k=headc(ptrc),headc(ptrc+1)-1
		     index=iabs(lit(k))
                     xindex=x(index)
		     if (lit(k).lt.0) then
			  y=-1
                     else
			  y=1
                     endif
		     if (index.eq.i) then
		          p0=xindex*y
		          p1=-xindex*y
                     else
		          p0=xindex*y
		          p1=xindex*y
		     endif
		     if (scr0.eq.0) then
			  if (p0.eq.1) scr0=1
                     endif
		     if (scr1.eq.0) then
			  if (p1.eq.1) scr1=1
                     endif
20              continue
		if(scr1.gt.scr0) then
		     gainx(i)=gainx(i)+w(ptrc)
                endif
		if(scr0.gt.scr1) then
		     gainx(i)=gainx(i)-w(ptrc)
                endif
		ptr=nextp(ptr)
		goto 10
30         continue

	   ptr=headm(i)
40         if (ptr.eq.0) goto 60
		scr0=0
		scr1=0
		ptrc=clause(ptr)
		do 50 k=headc(ptrc),headc(ptrc+1)-1
		     index=iabs(lit(k))
                     xindex=x(index)
		     if (lit(k).lt.0) then
			  y=-1
                     else
			  y=1
                     endif
		     if (index.eq.i) then
		          p0=xindex*y
		          p1=-xindex*y
                     else
		          p0=xindex*y
		          p1=xindex*y
		     endif
		     if (scr0.eq.0) then
			  if (p0.eq.1) scr0=1
                     endif
		     if (scr1.eq.0) then
			  if (p1.eq.1) scr1=1
                     endif
50              continue

		if(scr1.gt.scr0) then
		     gainx(i)=gainx(i)+w(ptrc)
                endif
		if(scr0.gt.scr1) then
		     gainx(i)=gainx(i)-w(ptrc)
                endif
		ptr=nextm(ptr)
		goto 40
60         continue
           if (gainx(i).gt.lgain) then
              lgain=gainx(i)
              li=i
           endif
70    continue
      if (lgain.gt.0) then
         x(li)=-x(li)
         smgain=smgain+lgain
      endif

80    lgain=0
      do 150 l=xptr(li), xptr(li+1)-1
          i=xx(l)
          gainx(i)=0
          ptr=headp(i)
90        if (ptr.eq.0) goto 110
             scr0=0
             scr1=0
	     ptrc=clause(ptr)
             do 100 k=headc(ptrc),headc(ptrc+1)-1
                index=iabs(lit(k))
                xindex=x(index)
                if (lit(k).lt.0) then
                   y=-1
                else
                   y=1
                endif
                if (index.eq.i) then
                   p0=xindex*y
                   p1=-xindex*y
                else
                   p0=xindex*y
                   p1=xindex*y
                endif
                if (scr0.eq.0) then
                   if (p0.eq.1) scr0=1
                endif
                if (scr1.eq.0) then
                   if (p1.eq.1) scr1=1
                endif
100         continue
             if(scr1.gt.scr0) then
                gainx(i)=gainx(i)+w(ptrc)
             endif
             if(scr0.gt.scr1) then
                gainx(i)=gainx(i)-w(ptrc)
             endif
             ptr=nextp(ptr)
             goto 90
110     continue

        ptr=headm(i)
120     if(ptr.eq.0) goto 140
               scr0=0
               scr1=0
               ptrc=clause(ptr)
               do 130 k=headc(ptrc),headc(ptrc+1)-1
                      index=iabs(lit(k))
                      xindex=x(index)
                      if (lit(k).lt.0) then
                         y=-1
                      else
                         y=1
                      endif
                      if (index.eq.i) then
                         p0=xindex*y
                         p1=-xindex*y
                      else
                         p0=xindex*y
                         p1=xindex*y
                      endif
                      if (scr0.eq.0) then
                         if (p0.eq.1) scr0=1
                      endif
                      if (scr1.eq.0) then
                         if (p1.eq.1) scr1=1
                      endif
130            continue
               if(scr1.gt.scr0) then
                      gainx(i)=gainx(i)+w(ptrc)
               endif
               if(scr0.gt.scr1) then
                      gainx(i)=gainx(i)-w(ptrc)
               endif
               ptr=nextm(ptr)
               goto 120
140       continue
150   continue
c     -------------------------------------------------------------
c     Find largest gain(i)
c     -------------------------------------------------------------
      do 160 ii=1,n
       if (gainx(ii).gt.lgain) then
                lgain=gainx(ii)
                li=ii
       endif
160   continue

      if (lgain.gt.0) then
           x(li)=-x(li)
           smgain=smgain+lgain
           goto 80
      else
           wsat=wsat+smgain
           return
      endif
c     ------------------------------------------------------------------
      end




      subroutine pr( clause, gainx,  headc,  headm,  headp,  lit,
     +               m,      mp1,    n,      nextm,  nextp,  np1,
     +               ntn,    numlit, w,      wsat,   x,      xptr,
     +               xx ,
     +               nelite, elite, maxelt, seede, symdif, msymdf,
     +               mxdpth, ypr,  bstelt, wstelt, optxpr, bestpr )
c     ------------------------------------------------------------------
c
c     pr: mixed path-relinikng 
c
c     ------------------------------------------------------------------
      implicit none

c     ------------------------------------------------------------------
c
c     Integer input parameters:
c
c        m              - number of clauses
c        mp1            - m + 1
c        n              - number of variables
c        np1            - n + 1
c        ntn            - n * n
c        numlit         - number of literals
c
c     ------------------------------------------------------------------
      integer  m,      mp1,    n,      np1,    ntn,    numlit

c     ------------------------------------------------------------------
c
c     Integer input/output parameter:
c
c        wsat           - weight of assignment
c
c     ------------------------------------------------------------------
      integer  wsat,wsatg

c     ------------------------------------------------------------------
c
c     Integer local parameters:
c
c        i              - literal index
c        ii             - do loop parameter
c        index          - literal index
c        k              - do loop parameter
c        l              - do loop parameter
c        lgain          - value of largest gain
c        li             - index of largest gain
c        p0             - xindex*y
c        p1             - -xindex*y
c        ptr             - pointer
c        ptrc            - clause(ptr)
c        scr0            - indicator
c        scr1            - indicator
c        smgain          - sum of gains
c        xindex          - x(index)
c        y               - indicator
c
c     ------------------------------------------------------------------
      integer  i,      ii,     index,  k,      l,      lgain,  li,
     +         p0,     p1,     ptr,    ptrc,   scr0,   scr1,    smgain,
     +         xindex, y

c     ------------------------------------------------------------------
c
c     Integer input arrays:
c
c        clause(mlit)   - clause which literal belongs to
c        headc(np1)     - pointer to first literal in clause
c        headm(nmax)    - pointer to first negated literal of variable
c        headp(nmax)    - pointer to first unnegated literal of variable
c        lit(mlit)      - variable of literal (>0=unnegated, <0=negated)
c        nextm(mlit)    - pointer to next minus clause
c        nextp(mlit)    - pointer to next plus clause
c        w(mmax)        - clause weight
c        xptr(mmax)     - pointer to first element of clause in 
c                         relationship list
c        xx(ntnmax)     - relationship list
c
c     ------------------------------------------------------------------
      integer  clause(numlit), headc(mp1),     headm(n),
     +         headp(n),       lit(numlit),    nextm(numlit),
     +         nextp(numlit),  w(m),           xptr(np1),
     +         xx(ntn)

c     ------------------------------------------------------------------
c
c     Integer input/output array:
c
c        x(nmax)        - assignment
c
c     ------------------------------------------------------------------
      integer  x(n)

c     ------------------------------------------------------------------
c
c     Integer local array:
c
c        gainx(nmax)    - gain achieved by flipping assignment
c
c     ------------------------------------------------------------------
      integer  gainx(n)

      integer  maxelt
      integer  nelite, elite(maxelt*(n+1)), seede, symdif(n), 
     +         msymdf, mxdpth, ypr(n), bstelt, wstelt, minsd, sd
      integer  iminsd,j,nsd,iii,bestpr,optxpr(n),depth,maxnsd,symcut

c     ------------------------------------------------------------------
c     check if x should be inserted into elite set
c     ------------------------------------------------------------------
      if (nelite .lt. maxelt) then
           do 77701 i=1,n
                elite(nelite*(n+1)+i)=x(i)
77701      continue
           elite(nelite*(n+1)+n+1)=wsat
           nelite=nelite+1
           if (wsat .lt. wstelt) then
                wstelt=wsat
           endif
           if (wsat .gt. bstelt) then
                bstelt=wsat
           endif
c          print elite set
c          print *,'nelite = ',nelite
c          print *,'bstelt, wstelt = ',bstelt, wstelt
c           do 77706 i=1,nelite
c                print *,'elite:', (elite((i-1)*(n+1)+j),j=1,n+1)
c77706      continue

c     ------------------------------------------------------------------
           return
      endif

c     elite set full

      minsd=n

c     first case: solution is better than best elite

      if (wsat .gt. bstelt) then
c          find who leaves ES
           bstelt=wsat
           do 77702 i=1,nelite
               sd=0
               do 77703 j=1,n
                    index= (nelite-1)*n+j
                    if (x(j) .ne. elite(index)) sd=sd+1
77703          continue
               if (sd .lt. minsd) then
                    minsd=sd
                    iminsd=i
               endif
77702      continue
c          print *,' most similar: minsd = ',minsd
c          make most similar sol leave
           do 77704 j=1,n
               index= (iminsd-1)*(n+1)+j
               elite(index) = x(j)
77704      continue
           if ( elite(index+1) .eq. wstelt) then
c               this part can be improved with a heap
                wstelt=2147483647
                do 77705 j=1,nelite
                     if (j.ne.iminsd) then
                          i= (j-1)*(n+1)+n+1
                          if (elite(i) .lt. wstelt) then
                               wstelt=elite(i)
                          endif
                     endif
77705           continue
           endif
           elite(index+1) = wsat
      endif

c     second case: solution is better than worst, but worse than 
c     best

      if ((wsat .le. bstelt) .and. (wsat .gt. wstelt)) then

c          check if x is suff. diff. from all elite solns

           symcut=msymdf*n/100
           do 17702 i=1,nelite
               sd=0
               do 17703 j=1,n
                    index= (nelite-1)*n+j
                    if (x(j) .ne. elite(index)) sd=sd+1
17703          continue
               if (sd .lt. symcut) goto 17709
17702      continue

c          find who leaves

           minsd=n

           do 27702 i=1,nelite
	       if (elite((nelite-1)*n+n+1) .le. wsat) then
                  sd=0
                  do 27703 j=1,n
                       index= (nelite-1)*n+j
                       if (x(j) .ne. elite(index)) sd=sd+1
27703             continue
                  if (sd .lt. minsd) then
                       minsd=sd
                       iminsd=i
                  endif
               endif
27702      continue

c          replace elite solution with new solution

           do 17704 j=1,n
               index= (iminsd-1)*(n+1)+j
               elite(index) = x(j)
17704      continue
           if ( elite(index+1) .eq. wstelt) then
c               this part can be improved with a heap
                wstelt=2147483647
                do 17705 j=1,nelite
                     if (j.ne.iminsd) then
                          i= (j-1)*(n+1)+n+1
                          if (elite(i) .lt. wstelt) then
                               wstelt=elite(i)
                          endif
                     endif
17705           continue
           endif
           elite(index+1) = wsat
      endif
17709 continue

c     print elite set
c     print *,'nelite = ',nelite
c     print *,'bstelt, wstelt = ',bstelt, wstelt
c     do 77709 i=1,nelite
c          print *,'elite:', (elite((i-1)*(n+1)+j),j=1,n+1)
c77709 continue

c     select solution (ypr) at random from PR to link with x

      call randp(seede)
      index=(seede/(2147483647/nelite))+1

c     copy elite solution into ypr

      do 77801 i=1,n
	   ypr(i)=x(i)
           x(i) = elite((index-1)*(n+1)+i)
77801 continue
      wsatg=wsat
      wsat=elite((index-1)*(n+1)+n+1)

c     compute symmetric difference between x and xpr

      nsd=0
      do 77802 i=1,n
           if (x(i) .ne. ypr(i)) then
                nsd=nsd+1
                symdif(nsd)=i
           endif
77802 continue

      bestpr=max0(wsat, wsatg)
c     print *,' PR: xonj, yobj = ', wsat, elite((index-1)*(n+1)+n+1)
      maxnsd=nsd
      depth=0
88801 if (nsd .eq. 0) goto 88802

      smgain=0
      lgain= -2147483647
      do 70 iii=1,nsd
           i = symdif(iii)
           gainx(i)=0
	   ptr=headp(i)
10         if(ptr.eq.0) goto 30
		scr0=0
		scr1=0
		ptrc=clause(ptr)
		do 20 k=headc(ptrc),headc(ptrc+1)-1
		     index=iabs(lit(k))
                     xindex=x(index)
		     if (lit(k).lt.0) then
			  y=-1
                     else
			  y=1
                     endif
		     if (index.eq.i) then
		          p0=xindex*y
		          p1=-xindex*y
                     else
		          p0=xindex*y
		          p1=xindex*y
		     endif
		     if (scr0.eq.0) then
			  if (p0.eq.1) scr0=1
                     endif
		     if (scr1.eq.0) then
			  if (p1.eq.1) scr1=1
                     endif
20              continue
		if(scr1.gt.scr0) then
		     gainx(i)=gainx(i)+w(ptrc)
                endif
		if(scr0.gt.scr1) then
		     gainx(i)=gainx(i)-w(ptrc)
                endif
		ptr=nextp(ptr)
		goto 10
30         continue

	   ptr=headm(i)
40         if (ptr.eq.0) goto 60
		scr0=0
		scr1=0
		ptrc=clause(ptr)
		do 50 k=headc(ptrc),headc(ptrc+1)-1
		     index=iabs(lit(k))
                     xindex=x(index)
		     if (lit(k).lt.0) then
			  y=-1
                     else
			  y=1
                     endif
		     if (index.eq.i) then
		          p0=xindex*y
		          p1=-xindex*y
                     else
		          p0=xindex*y
		          p1=xindex*y
		     endif
		     if (scr0.eq.0) then
			  if (p0.eq.1) scr0=1
                     endif
		     if (scr1.eq.0) then
			  if (p1.eq.1) scr1=1
                     endif
50              continue

		if(scr1.gt.scr0) then
		     gainx(i)=gainx(i)+w(ptrc)
                endif
		if(scr0.gt.scr1) then
		     gainx(i)=gainx(i)-w(ptrc)
                endif
		ptr=nextm(ptr)
		goto 40
60         continue
           if (gainx(i).gt.lgain) then
              lgain=gainx(i)
              li=i
           endif
70    continue

c     -------------------------------------------------------------
c     Find largest gain(i)
c     -------------------------------------------------------------
      do 160 iii=1,nsd
          ii = symdif(iii)
          if (gainx(ii).gt.lgain) then
                lgain=gainx(ii)
                li=ii
          endif
160   continue

      x(li)=-x(li)
      smgain=smgain+lgain
      wsat=wsat+smgain
       
c     check if x is best so far in PR
      depth=depth+1
      if (wsat.gt.bestpr) then
            bestpr=wsat
            do 66601 i=1,n
                 optxpr(i)=x(i)
66601       continue
c           print *,'depth of best pr = ',depth,' of maxnsd = ',maxnsd
      endif
c     remove li from symdif
      do 66603 i=1,nsd
         if (symdif(i) .eq. li) then
              symdif(i)=symdif(nsd)
              nsd=nsd-1
              goto 88801
         endif
66603 continue
88802 continue
c     print *,'end of PR: bestpr =  ',bestpr

c     try to insert optxpr into elite set

      minsd=n
      if (bestpr .gt. bstelt) then
c          find who leaves ES
           bstelt=bestpr
           do 47702 i=1,nelite
               sd=0
               do 47703 j=1,n
                    index= (nelite-1)*n+j
                    if (optxpr(j) .ne. elite(index)) sd=sd+1
47703          continue
               if (sd .lt. minsd) then
                    minsd=sd
                    iminsd=i
               endif
47702      continue
c          print *,' most similar: minsd = ',minsd
c          make most similar sol leave
           do 47704 j=1,n
               index= (iminsd-1)*(n+1)+j
               elite(index) = optxpr(j)
47704      continue
           if ( elite(index+1) .eq. wstelt) then
c               this part can be improved with a heap
                wstelt=2147483647
                do 47705 j=1,nelite
                     if (j.ne.iminsd) then
                          i= (j-1)*(n+1)+n+1
                          if (elite(i) .lt. wstelt) then
                               wstelt=elite(i)
                          endif
                     endif
47705           continue
           endif
           elite(index+1) = bestpr
      endif

c     second case: solution is better than worst, but worse than 
c     best

      if ((bestpr .le. bstelt) .and. (bestpr .gt. wstelt)) then

c          check if x is suff. diff. from all elite solns

           symcut=msymdf*n/100
           do 33302 i=1,nelite
               sd=0
               do 33303 j=1,n
                    index= (nelite-1)*n+j
                    if (optxpr(j) .ne. elite(index)) sd=sd+1
33303          continue
               if (sd .lt. symcut) goto 33309
33302      continue

c          find who leaves

           minsd=n

           do 44402 i=1,nelite
	       if (elite((nelite-1)*n+n+1) .le. bestpr) then
                  sd=0
                  do 44403 j=1,n
                       index= (nelite-1)*n+j
                       if (optxpr(j) .ne. elite(index)) sd=sd+1
44403             continue
                  if (sd .lt. minsd) then
                       minsd=sd
                       iminsd=i
                  endif
               endif
44402      continue

c          replace elite solution with new solution

           do 44404 j=1,n
               index= (iminsd-1)*(n+1)+j
               elite(index) = optxpr(j)
44404      continue
           if ( elite(index+1) .eq. wstelt) then
c               this part can be improved with a heap
                wstelt=2147483647
                do 44405 j=1,nelite
                     if (j.ne.iminsd) then
                          i= (j-1)*(n+1)+n+1
                          if (elite(i) .lt. wstelt) then
                               wstelt=elite(i)
                          endif
                     endif
44405           continue
           endif
           elite(index+1) = bestpr
      endif
33309 continue

      return
c     ------------------------------------------------------------------
      end








      subroutine savsol( bestv,  bests,  n,      seed0,  wsat,   x,
     +                   xopt )
c     ------------------------------------------------------------------
c
c     savsol:  Saves current best solution in (bestv, xopt).
c
c     ------------------------------------------------------------------
      implicit none

c     ------------------------------------------------------------------
c
c     Integer input parameters:
c
c        n              - number of variables
c        seed0          - seed at start of current GRASP iteration
c        wsat           - weight of assignment
c
c     ------------------------------------------------------------------
      integer  n,      seed0,  wsat

c     ------------------------------------------------------------------
c
c     Integer output parameter:
c
c        bestv          - value of best truth assignment
c        bests          - seed at start of best GRASP iteration
c
c     ------------------------------------------------------------------
      integer  bestv,  bests

c     ------------------------------------------------------------------
c
c     Integer input array:
c
c        x(n)           - assignment
c
c     ------------------------------------------------------------------
      integer x(n)

c     ------------------------------------------------------------------
c
c     Integer output array:
c
c        xopt(n)        - best assignment
c
c     ------------------------------------------------------------------
      integer xopt(n)

c     ------------------------------------------------------------------------
      call copyi4( n,      x,   xopt)
      bestv=wsat
      bests=seed0
c     ------------------------------------------------------------------------
      return
      end








      subroutine copyi4(n,x,y)
c     ------------------------------------------------------------------
c
c     copyi4:  copies integer*4 array x of size n to integer*4 array y
c
c     ------------------------------------------------------------------
      implicit none
c     ------------------------------------------------------------------
c
c     Integer input parameter:
c
c        n              - size of array
c
c     ------------------------------------------------------------------
      integer n
c     ------------------------------------------------------------------
c
c     Integer local parameter:
c
c        i              - do loop parameter
c
c     ------------------------------------------------------------------
      integer i

c     ------------------------------------------------------------------
c
c     Integer input arrays:
c
c        x(n)           - input integer array
c
c     ------------------------------------------------------------------
      integer x(n)
c     ------------------------------------------------------------------
c
c     Integer output arrays:
c
c        y(n)           - output integer array
c
c     ------------------------------------------------------------------
      integer y(n)

      do 10 i=1,n
	   y(i)=x(i)
10    continue
c     ------------------------------------------------------------------
      return
      end
